import os
import pickle
import itertools

import numpy as np
import scipy.stats
import scipy.optimize
import scipy.interpolate
import sklearn.mixture

import handygenome
import handygenome.tools as tools
import handygenome.cnv.baf as libbaf
import handygenome.cnv.misc as cnvmisc


BAFCORRECT_PATH = os.path.join(handygenome.USERDATA_DIR, f'baf_correction_func.pickle')

def save_bafcorrect_func(func):
    with open(BAFCORRECT_PATH, 'wb') as outfile:
        pickle.dump(func, outfile)


def load_bafcorrect_func(x_cutoff=None):
    with open(BAFCORRECT_PATH, 'rb') as infile:
        func = pickle.load(infile)

    return winsorize_pointfive(func, x_cutoff=x_cutoff)

#

def save_bafcorrect_func_xcutoff(func, x_cutoff):
    with open(BAFCORRECT_PATH, 'wb') as outfile:
        pickle.dump((func, x_cutoff), outfile)

    
def load_bafcorrect_func_xcutoff():
    with open(BAFCORRECT_PATH, 'rb') as infile:
        func, x_cutoff = pickle.load(infile)

    return winsorize_pointfive(func, x_cutoff)


def winsorize_pointfive(func, x_cutoff=None):
    def newfunc(x):
        result = func(x)
        result[result > 0.5] = 0.5
        if x_cutoff is not None:
            result[x > x_cutoff] = 0.5
        return result

    return newfunc


########################
# baf simulation - new #
########################

def make_vafs_and_bafs(allele_portions, mean_depth, size, alpha_sum=100, dirichlet=True):
    """Help:
        - "allele_portions" is first normalized such that it sums to 1
        - The order of "allele_portions" is not important
        - Larger alpha_sum leads to less variation
        - Total depths are generated from poisson distribution
        - If "dirichlet" is False, plain multinomial distribution is used.
    """
    # normalize allele_portions such that sum becomes 1
    allele_portions = np.asarray(allele_portions)
    assert allele_portions.ndim == 1
    allele_portions = allele_portions / allele_portions.sum()

    # make total depths
    rng = np.random.default_rng()
    total_depths = make_simulated_total_depths(mean_depth, size, rng=rng)

    if dirichlet:
        dirichlet_alpha = allele_portions * alpha_sum
        allele_depths = tools.dirichlet_multinomial_rvs(total_depths, dirichlet_alpha, rng=rng)
    else:
        allele_depths = rng.multinomial(n=total_depths, pvals=allele_portions)

    # make vafs and bafs
    vafs = allele_depths / total_depths[:, np.newaxis]
    bafs = libbaf.get_baf_from_vaf(vafs)

    return {
        'total_depth': total_depths,
        'allele_depth': allele_depths,
        'vaf': vafs,
        'baf': bafs,
    }


def make_simulated_total_depths(mean_depth, N, rng=None):
    """Only non-zero values"""

    if rng is None:
        rng = np.random.default_rng()

    result = list()
    n = N
    while True:
        arr = rng.poisson.rvs(lam=mean_depth, size=n)
        #arr = scipy.stats.poisson.rvs(mean_depth, size=n)
        arr = arr[arr > 0]
        result.extend(arr)
        if len(result) == N:
            break
        else:
            n = N - len(result)
            continue
    return np.asarray(result)


def get_baf_statistic(baf_data, method='density'):
    assert method in ['density', 'gm', 'mean']

    baf_data = np.asarray(baf_data)
    assert baf_data.ndim == 1

    if method == 'density':
        density = scipy.stats.gaussian_kde(baf_data)
        argmin = scipy.optimize.minimize_scalar(lambda x: -density(x)[0], bounds=(0, 0.5))
        return argmin.x
    elif method == 'gm':  # gaussian mixture
        vafs = np.concatenate([baf_data, 1 - baf_data], axis=0)
        vafs = np.expand_dims(vafs, axis=1)
        gm = sklearn.mixture.GaussianMixture(n_components=2).fit(vafs)
        means = sorted(gm.means_.ravel())
        return (means[0] + (1 - means[1])) / 2
    elif method == 'mean':
        bafs = bafs[bafs > 0]
        return bafs.mean()


##################
# baf simulation #
##################

def make_vafs_bafs(baf, mean_depth, size):
    halfsize = int(size / 2)
    depths = scipy.stats.poisson.rvs(mu=mean_depth, size=size)
    alts = np.concatenate(
        (
            scipy.stats.binom.rvs(n=depths[:halfsize], p=baf), 
            scipy.stats.binom.rvs(n=depths[halfsize:], p=(1 - baf)), 
        )
    )

    vafs = alts / depths
    np.random.shuffle(vafs)

    bafs = vafs.copy()
    bafs[bafs > 0.5] = 1 - bafs[bafs > 0.5]

    return vafs, bafs


def make_simulated_germline_vafs(mean_depth, size, mode, a=None, b=None):
    depths = scipy.stats.poisson.rvs(mu=mean_depth, size=size)

    if mode == 'betabinom':
        alts = scipy.stats.betabinom.rvs(n=depths, a=a, b=b)
    elif mode == 'binom':
        alts = scipy.stats.binom.rvs(n=depths, p=0.5)

    vafs = alts / depths

    return vafs


def make_simulated_germline_vafs_normal(size, std=0.1):
    result = scipy.stats.norm.rvs(loc=0.5, scale=std, size=size)
    result[result < 0] = 0
    result[result > 1] = 1
    return result


def infer_baf_minimize(baf_values, density=None):
    if density is None:
        density = scipy.stats.gaussian_kde(baf_values)

    argmin = scipy.optimize.minimize_scalar(lambda x: -density(x)[0], bounds=(0, 0.5))
    return argmin.x


def infer_baf_mixture(vaf_values):
    vals = vaf_values.reshape(-1, 1)
    gm = sklearn.mixture.GaussianMixture(n_components=2).fit(vals)
    means = sorted(gm.means_.ravel())
    return (means[0] + (1 - means[1])) / 2


def make_simulation_data(mean_depth=34, size=50, reps=100, true_bafs=None):
    if true_bafs is None:
        true_bafs = np.round(np.arange(0.01, 0.51, 0.01), 2)

    skew_data = dict()
    ebafmean_data = dict()
    ebafmin_data = dict()
    ebafmix_data = dict()
    baf_data = dict()
    vaf_data = dict()

    for baf in true_bafs:
        print(baf)

        skew_values = list()
        ebafmean_values = list()
        ebafmin_values = list()
        ebafmix_values = list()
        vafs_list = list()
        bafs_list = list()

        for _ in range(reps):
            vafs, bafs = make_vafs_bafs(baf=baf, mean_depth=mean_depth, size=size)
            vafs_list.append(vafs)
            bafs_list.append(bafs)
            skew_values.append(scipy.stats.skew(bafs))
            ebafmean_values.append(bafs.mean())
            ebafmin_values.append(infer_baf_minimize(bafs))
            ebafmix_values.append(infer_baf_mixture(vafs))

        skew_data[baf] = np.array(skew_values)
        ebafmean_data[baf] = np.array(ebafmean_values)
        ebafmin_data[baf] = np.array(ebafmin_values)
        ebafmix_data[baf] = np.array(ebafmix_values)
        baf_data[baf] = np.array(bafs_list)
        vaf_data[baf] = np.array(vafs_list)

    return {
        'skew': skew_data,
        'mean': ebafmean_data,
        'minimize': ebafmin_data,
        'mixture': ebafmix_data,
        'baf': baf_data,
        'vaf': vaf_data,
    }


def get_xs_ratio(data, realdata, size=10000, rep=10000):
    realdata_means = np.array([np.random.choice(realdata, size).mean() for x in range(rep)])
    xs_ratio = realdata_means.mean() / data['mean'][0.5].mean()
    return xs_ratio


def get_x_cutoff(realdata, size=100, rep=10000, q=0.05):
    realdata_means = np.array([np.random.choice(realdata, size).mean() for x in range(rep)])
    return np.quantile(realdata_means, q)


def xs_ratio_correction(xs_ratio, xs, ys):
    if xs_ratio is None:
        corrected_xs = None
        interp = scipy.interpolate.interp1d(xs, ys, fill_value='extrapolate')
    else:
        corrected_xs = xs * xs_ratio
        interp = scipy.interpolate.interp1d(corrected_xs, ys, fill_value='extrapolate')

    return interp, corrected_xs


def make_bafcorrector_type1(data, xs_ratio=None):
    xs = list()
    ys = list()
    for true_baf_val, ebaf_vals in data['mean'].items():
        xs.append(ebaf_vals.mean())
        ys.append(true_baf_val)

    xs = np.array(xs)
    ys = np.array(ys)

    interp, corrected_xs = xs_ratio_correction(xs_ratio, xs, ys)
    return interp, xs, ys, corrected_xs


def make_bafcorrector_type2(data, xs_ratio=None):
    xs = list()
    ys = list()
    for true_baf_val, ebaf_vals in data['mean'].items():
        xs.extend(ebaf_vals)
        ys.extend(itertools.repeat(true_baf_val, len(ebaf_vals)))

    xs = np.array(xs)
    ys = np.array(ys)

    interp, corrected_xs = xs_ratio_correction(xs_ratio, xs, ys)

    return interp, xs, ys, corrected_xs


def make_bafcorrector_from_realdata(data, realdata, mode='type1'):
    xs_ratio = get_xs_ratio(data, realdata, size=10000, rep=10000)
    x_cutoff = get_x_cutoff(realdata, size=100, rep=10000, q=0.05)
    if mode == 'type1':
        interp, xs, ys, corrected_xs = make_bafcorrector_type1(data, xs_ratio=xs_ratio)
    elif mode == 'type2':
        interp, xs, ys, corrected_xs = make_bafcorrector_type2(data, xs_ratio=xs_ratio)

    bafcorrector = winsorize_pointfive(interp, x_cutoff)

    return bafcorrector, interp, x_cutoff

        
# baf simulation v2 (230525)

#def make_simulated_total_depths(mean_depth, N):
#    result = list()
#    n = N
#    while True:
#        arr = scipy.stats.poisson.rvs(mean_depth, size=n)
#        arr = arr[arr > 0]
#        result.extend(arr)
#        if len(result) == N:
#            break
#        else:
#            n = N - len(result)
#            continue
#    return np.asarray(result)


def make_simulated_het_bafs(vaf, mean_depth, N):
    td = make_simulated_total_depths(mean_depth, N)
    ad = scipy.stats.binom.rvs(n=td, p=vaf)
    vafs = ad / td
    bafs = np.where(vafs > 0.5, 1 - vafs, vafs)
    return bafs


def make_simulated_hom_bafs(p_error, mean_depth, N):
    td = make_simulated_total_depths(mean_depth, N)
    ad = scipy.stats.binom.rvs(n=td, p=p_error)
    vafs = ad / td
    bafs = np.where(vafs > 0.5, 1 - vafs, vafs)
    return bafs


def make_simulated_germline_call_bafs(mean_depth, vaf, N, p_error, hom_portion):
    N_hom = int(N * hom_portion)
    N_het = N - N_hom
    return np.concatenate([
        make_simulated_het_bafs(vaf, mean_depth, N_het),
        make_simulated_hom_bafs(p_error, mean_depth, N_hom),
    ])


def infer_baf_density(bafs, bw, rmzero=True):
    if rmzero:
        bafs = bafs[bafs > 0]
    assert len(bafs) > 0
    if len(bafs) == 1:
        return bafs[0]
    else:
        peak_values, peak_densities, density = cnvmisc.get_density_peaks(bafs, bw_method=bw)
        #return np.average(peak_values, weights=peak_densities)
        if peak_values is None:
            return np.nan
        else:
            return max(peak_values)


def infer_baf_mean(bafs):
    bafs = bafs[bafs > 0]
    return bafs.mean()


def make_simulated_baf_dataset(mean_depth=30, N=10000, p_error=0.001, reps=10, hom_portion=0.6, verbose=False):
    true_bafs = np.arange(0, 0.51, 0.01)
    simbaf_data = list()
    for bafval in true_bafs:
        if verbose:
            print(f'baf value {bafval}')
        subdata = list()
        for _ in range(reps):
            #print(f'repetition {_}')
            simulated_bafs = make_simulated_germline_call_bafs(
                mean_depth, bafval, N, p_error, hom_portion,
            )
            subdata.append(simulated_bafs)
        simbaf_data.append(np.array(subdata))

    simbaf_data = np.stack(simbaf_data, axis=0)

    return simbaf_data


def make_inferred_baf_dataset(simbaf_data, bw=1, verbose=False, infer_method='density'):
    inferred_baf_data = list()
    for truebaf_group in simbaf_data:
        subdata = list()
        for bafvals in truebaf_group:
            if infer_method == 'density':
                inferred_baf = infer_baf_density(bafvals, bw=bw)
            elif infer_method == 'mean':
                inferred_baf = infer_baf_mean(bafvals)
            subdata.append(inferred_baf)
        inferred_baf_data.append(np.array(subdata))
    return np.stack(inferred_baf_data, axis=0)


def find_best_bandwidth(mean_depth):
    true_bafs = np.arange(0, 0.51, 0.01)
    simbaf_data = make_simulated_baf_dataset(
        mean_depth=mean_depth, N=1000, p_error=0.001, reps=10, hom_portion=0.6, verbose=False
    )
    def target(bw):
        inferred_baf_data = make_inferred_baf_dataset(simbaf_data, bw=bw, verbose=False, infer_method='density')
        stdsum = np.std(inferred_baf_data, axis=1).sum()
        means = np.mean(inferred_baf_data, axis=1)
        devsum = np.sqrt(np.sum((true_bafs - means) ** 2))
        return stdsum + devsum

    bwlist = np.arange(0.5, 1, 0.05)
    targetvals = [target(x) for x in bwlist]
    return bwlist[np.argmin(targetvals)]





"""This version abandons gc data saving with bigwig and saves DataFrame as tsv file"""

import os
import itertools
import functools

import Bio.SeqUtils
import numpy as np
import pandas as pd
import pyranges as pr

import handygenome
import handygenome.refgenome.refgenome as refgenome
import handygenome.logutils as logutils
import handygenome.pyranges_helper as pyranges_helper
import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome


GCDATA_DIR = os.path.join(handygenome.USERDATA_DIR, 'gcdata')
os.makedirs(GCDATA_DIR, exist_ok=True)
#if not os.path.exists(GCDATA_DIR):
#    os.mkdir(GCDATA_DIR)


# making and saving gc fraction bigwig files

def write_gcfile(outfile_path, fasta, binsize=100):
    chromdict = refgenome.ChromDict(fasta=fasta)
    bin_gr = chromdict.to_gr().window(binsize)
    gc_df = bin_gr.df
    gc_df['GC'] = calculate_gcvals(
        tuple(bin_gr.Chromosome), 
        tuple(bin_gr.Start), 
        tuple(bin_gr.End), 
        fasta, 
        window=None, 
        as_array=True,
    )
    gc_df.to_csv(outfile_path, sep='\t', header=True, index=False)


def get_gcfile_path(refver, binsize):
    refver = refgenome.RefverDict.standardize(refver)
    gcdata_refver_dir = os.path.join(GCDATA_DIR, refver)
    if not os.path.exists(gcdata_refver_dir):
        os.mkdir(gcdata_refver_dir)
    return os.path.join(gcdata_refver_dir, f'gc_binsize_{100}.tsv.gz')


@functools.cache
def get_gc_df(refver, binsize, coords_as_index=True):
    gcfile_path = get_gcfile_path(refver, binsize)

    if not os.path.exists(gcfile_path):
        logutils.print_timestamp(f'There is no pre-existing gc data file. A new one is being created. It may take a few minutes.')
        fasta = refgenome.get_fasta(refver)
        write_gcfile(gcfile_path, fasta, binsize=binsize)
        logutils.print_timestamp(f'Finished making a gc data file.')

    if coords_as_index:
        index_col = ['Chromosome', 'Start', 'End']
    else:
        index_col = None

    return pd.read_csv(
        gcfile_path, 
        sep='\t', 
        header=0, 
        dtype={'Chromosome': 'category', 'Start': int, 'End': int, 'GC': float},
        index_col=index_col,
    )


###########################################

# getting array of gc fraction values

def make_fetchargs_func(window, fasta):
    if window is None:
        def get_fetchargs(chrom, start0, end0):
            return chrom, start0, end0
    else:
        chromlens = dict(zip(fasta.references, fasta.lengths))
        pad_left = int(window / 2)
        if window % 2 == 0:
            pad_right = pad_left
        else:
            pad_right = pad_left + 1

        def get_fetchargs(chrom, start0, end0):
            if end0 - start0 >= window:
                return chrom, start0, end0
            else:
                mid = int((start0 + end0 - 1) / 2)
                new_start0 = max(0, mid - pad_left)
                new_end0 = min(chromlens[chrom], mid + pad_right)
                return chrom, new_start0, new_end0

    return get_fetchargs


def calculate_gcvals_generator(chroms, start0s, end0s, fasta, get_fetchargs):
    prev_chrom = ''
    for chrom, start0, end0 in zip(chroms, start0s, end0s):
        if chrom != prev_chrom:
            print(chrom)
        prev_chrom = chrom
        yield float(Bio.SeqUtils.gc_fraction(fasta.fetch(*get_fetchargs(chrom, start0, end0))))


@functools.cache
def calculate_gcvals(chroms, start0s, end0s, fasta, window=None, as_array=True):
    logutils.print_timestamp('Beginning gc fraction calculation')
    get_fetchargs = make_fetchargs_func(window, fasta)
    #val_gen = calculate_gcvals_generator(chroms, start0s, end0s, fasta, get_fetchargs)
    val_gen = (
        float(Bio.SeqUtils.gc_fraction(fasta.fetch(*get_fetchargs(chrom, start0, end0))))
        for chrom, start0, end0 in zip(chroms, start0s, end0s)
    )

    if as_array:
        return np.fromiter(val_gen, float)
    else:
        return list(val_gen)


def calculate_gcvals_with_df(df, fasta, window=None, as_array=True):
    return calculate_gcvals(
        tuple(df.Chromosome), 
        tuple(df.Start), 
        tuple(df.End), 
        fasta, 
        window=window, 
        as_array=as_array,
    )


def load_gcvals(chroms, start0s, end0s, refver, binsize, window=None, fasta=None):
    """Returns:
        A pandas Series
    """
    assert not ((window is not None) and (fasta is None))

    get_fetchargs = make_fetchargs_func(window, fasta)
    gc_df = get_gc_df(refver, binsize)
    left_gr = pr.PyRanges(chromosomes=chroms, starts=start0s, ends=end0s)
    joined_gr = pyranges_helper.join(left_gr, pr.PyRanges(gc_df), merge='weighted_mean', how='left')

    return joined_gr.GC


def load_gcvals_with_df(df, refver, binsize, fasta, window=None):
    return load_gcvals(
        df.Chromosome, df.Start, df.End, 
        refver=refver, 
        binsize=binsize, 
        fasta=refgenome.get_fasta(refver), 
        window=window,
    )


# add gc fraction values to a depth dataframe

@deco.get_deco_num_set_differently(('fasta', 'refver'), 1)
def add_gc_calculating(df, *, refver=None, fasta=None, window=None):
    """Args:
        df: pandas.DataFrame or pyranges.PyRanges
        refver, fasta: refver is recommended because with refver, default 
            fasta object with the given refver is loaded and subsequently
            used for "calculate_gcvals" function. Using a fasta object with
            a fixed id can make use of "cache" functionality of "calculate_gcvals".

    Modifies input DataFrame
    """
    if fasta is None:
        fasta = refgenome.get_fasta(refver)

    gcvals = calculate_gcvals(
        tuple(df.Chromosome), 
        tuple(df.Start), 
        tuple(df.End), 
        fasta, 
        window=window, 
        as_array=True,
    )

    if isinstance(df, pd.DataFrame):
        df['GC'] = gcvals
    elif isinstance(df, pr.PyRanges):
        df.GC = gcvals

    return df


def add_gc_loading(df, refver, binsize, window=None):
    """Args:
        df: pandas.DataFrame or pyranges.PyRanges
    Changes in-place
    """
    gcvals = load_gcvals(
        df.Chromosome, df.Start, df.End, 
        refver=refver, 
        binsize=binsize, 
        fasta=refgenome.get_fasta(refver), 
        window=window,
    )
    if isinstance(df, pd.DataFrame):
        df['GC'] = gcvals
    elif isinstance(df, pr.PyRanges):
        df.GC = gcvals





import numpy as np
#import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt


def make_quadmesh(len_x, len_y, width_x, width_y):
    xs = np.tile(np.arange(0, (len_x + 1) * width_x, width_x), (len_y + 1, 1))
    ys = np.tile(np.arange(0, (len_y + 1) * width_y, width_y), (len_x + 1, 1)).T
    coords = np.dstack((xs, ys))
    return mpl.collections.QuadMesh(coords, edgecolors='black', facecolors='none', linewidth=0.5)


def setup_main(ax, samples, variants):
    ax.set_aspect(1)
    
    ax.set(xlim=(0, 2 * len(samples)), ylim=(0, len(variants)))
    ax.add_collection(make_quadmesh(2 * len(samples), len(variants), 1, 1))

    ax.set_xticks(np.arange(1, 2 * len(samples), 2), samples, minor=False)
    ax.tick_params(axis='x', which='major', bottom=False, rotation=90, pad=10)
    ax.set_xticks(np.arange(0.5, 2 * len(samples), 1), np.tile(['w', 'p'], len(samples)), minor=True)
    ax.tick_params(axis='x', which='minor', bottom=False, rotation=0, pad=2, labelsize=8)

    ax.set_yticks(np.arange(0.5, len(variants), 1), variant_names, va='center')
    ax.tick_params(axis='y', left=False)
    

def setup_bottom(ax, samples, title):
    ax.set_aspect(1)
    
    ax.set(xlim=(0, 2 * len(samples)), ylim=(0, 1))
    ax.add_collection(make_quadmesh(len(samples), 1, 2, 1))
    ax.tick_params(axis='both', which='both', labelbottom=False, labelleft=False, bottom=False, left=False)
    ax.set_ylabel(title, rotation=0, va='center', ha='right')

    
def setup_left(ax, variants, title):
    ax.set_aspect(1)
    
    ax.set(xlim=(0, 1), ylim=(0, len(variants)))
    ax.add_collection(make_quadmesh(1, len_y, 1, 1))
    ax.tick_params(axis='both', which='both', labelbottom=False, labelleft=False, bottom=False, left=False)
    ax.set_xlabel(title, rotation=90, va='top', ha='center')


def write_bottom(fig, ax_values, ax_cb, values, cbar_min, cbar_max):
    mappable = mpl.cm.ScalarMappable(norm=mpl.colors.Normalize(vmin=cbar_min, vmax=cbar_max), cmap=mpl.cm.cool)
    fig.colorbar(mappable, cax=ax_cb, orientation='horizontal')
    
    for t, x in zip([str(k) for k in values], np.arange(1, 2 * len(values), 2)):
        ax_values.annotate(t, (x, 0.5), ha='center', va='center', size=8)
    
    colors = mappable.to_rgba(values)
    pcol = mpl.collections.PatchCollection(
        [
            mpl.patches.Rectangle((x, 0), 2, 1, linewidth=0, color=c)
            for x, c in zip(np.arange(0, 2 * len(values), 2), colors)
        ],
        match_original=True,
    ) 
    ax_values.add_collection(pcol)
    

def get_var_id(vp):
    if not hasattr(vp, 'oncokb'):
        vp.oncokb = liboncokb.query_hgvsg(vp.vcfspec.to_hgvsg(), ONCOKB_TOKEN)
    var_id = re.sub('^.+the (.+) mutation.+$', '\\1', vp.oncokb['variantSummary'])
    return var_id

    
def fill_main(ax, wgs_driver_vplists, panel_driver_vplists, SVinfo,  samples, variants):
    patches = list()
    var_id_pat = re.compile('^.+the (.+) mutation.+$')
    
    for sampleid, x in zip(samples, np.arange(0.5, len(samples) * 2, 2)):
        x_mid_wgs = x
        x_mid_panel = x + 1
        for var_name, y_bot in zip(variant_names, np.arange(0, len(variants), 1)):
            # FUSION
            if '-' in var_name:
                gene1, gene2 = var_name.split('-')
                for dic in SVinfo['wgs'][sampleid]:
                    if gene1 in dic['genes'] and gene2 in dic['genes'] and dic['strand_aligned']:
                        patches.append(mpl.patches.Rectangle((x_mid_wgs - 0.5, y_bot), 1, 1, color='r', alpha=0.5, zorder=-1))
                        break
                for dic in SVinfo['panel'][sampleid]:
                    if gene1 in dic['genes'] and gene2 in dic['genes'] and dic['strand_aligned']:
                        patches.append(mpl.patches.Rectangle((x_mid_panel - 0.5, y_bot), 1, 1, color='r', alpha=0.5, zorder=-1))
                        break
            else:
                if var_name.startswith('EGFR'):
                    if var_name == 'EGFR others':
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return ('EGFR' in var_id) and (var_id not in [    
                                'EGFR L858R',
                                'EGFR E746_A750del',
                                'EGFR A767_V769dup',])
                    else:
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return var_id == var_name
                elif var_name.startswith('KRAS'):
                    if var_name == 'KRAS others':
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return ('KRAS' in var_id) and (var_id not in [   
                                'KRAS G12C',
                                'KRAS G12D',
                                'KRAS G12V',                                
                            ])
                    else:
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return var_id == var_name
                elif var_name.startswith('BRAF'):
                    if var_name == 'BRAF others':
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return ('BRAF' in var_id) and (var_id not in [   
                                'BRAF G466V',
                            ])
                    else:
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return var_id == var_name                   
                else:
                    if var_name == 'others':
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return (
                                not any(x in var_id for x in ['EGFR', 'KRAS', 'BRAF',                             'PIK3CA', 
                            'PTEN',
                            'TP53', 
                            'KEAP1',
                            'ATM',]) 
                            )
                    else:
                        def vpmatcher(vp):
                            var_id = get_var_id(vp)
                            return var_name in var_id
                        
                hit = False
                for vp in wgs_driver_vplists[sampleid]:
                    if vpmatcher(vp):
                        hit = True
                        break
                if hit:
                    vaf = vp.get_vaf(f'{sampleid}_tumor')
                    if not hasattr(vp, 'ccfinfo'):
                        vp.ccfinfo = vp.get_ccf_CNm(segments_df=CNV_SEGMENTS[sampleid], is_female=IS_FEMALE[sampleid], cellularity=PREVIOUS_PURITY[sampleid], sampleid=f'{sampleid}_tumor')
                    ccf = vp.ccfinfo.ccf
                    
                    # print(vaf, ccf)
                    
                    ax.hlines(y=(y_bot + vaf), xmin=(x_mid_wgs - 0.5), xmax=(x_mid_wgs + 0.5), color='blue')
                    ax.hlines(y=(y_bot + ccf), xmin=(x_mid_wgs - 0.5), xmax=(x_mid_wgs + 0.5), color='orange')
                    
                hit = False
                for vp in panel_driver_vplists[sampleid]:
                    if vpmatcher(vp):
                        hit = True
                        break
                if hit:
                    vaf = vp.get_vaf(f'{sampleid}_panel')
                    if not hasattr(vp, 'ccfinfo'):
                        vp.ccfinfo = vp.get_ccf_CNm(segments_df=CNV_SEGMENTS[sampleid], is_female=IS_FEMALE[sampleid], cellularity=PREVIOUS_PURITY[sampleid], sampleid=f'{sampleid}_panel')
                    ccf = vp.ccfinfo.ccf
                    
                    # print(vaf, ccf)
                    
                    ax.hlines(y=(y_bot + vaf), xmin=(x_mid_panel - 0.5), xmax=(x_mid_panel + 0.5), color='blue')
                    ax.hlines(y=(y_bot + ccf), xmin=(x_mid_panel - 0.5), xmax=(x_mid_panel + 0.5), color='orange')

          
    ax.add_collection(mpl.collections.PatchCollection(patches, match_original=True))


def make_heatmap(variants, samples, purities, ploidies):
    fig = plt.figure(
        figsize=(20, 5), 
        layout='constrained',
    )

    axd = fig.subplot_mosaic(
        [
            ['lleft', 'left', 'main'],
            ['purity_cb', 'purity_cb', 'purity'],
            # ['.', '.', 'purity_cb'],
            ['ploidy_cb', 'ploidy_cb', 'ploidy'],
            # ['.', '.', 'ploidy_cb']
        ],
        width_ratios=(1, 1, len_x),
        height_ratios=(len_y, 1, 1),
    )

    setup_main(axd['main'], samples, variant_names)
    setup_bottom(axd['purity'], samples, 'purity')
    setup_bottom(axd['ploidy'], samples, 'ploidy')
    setup_left(axd['left'], variant_names, 'left1')
    setup_left(axd['lleft'], variant_names, 'left2')

    write_bottom(fig, axd['purity'], axd['purity_cb'], [purities[x] for x in samples], 0, 1)
    write_bottom(fig, axd['ploidy'], axd['ploidy_cb'], [ploidies[x] for x in samples], 1, 8)

    fill_main(axd['main'], wgs_driver_vplists, panel_driver_vplists, SVinfo, samples, variants)



    



import re
import textwrap
import sys
import os
import tempfile
import contextlib

import pysam

import handygenome.refgenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.variant.infoformat as infoformat
import handygenome.vcfeditor.headerhandler as headerhandler
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.variant.vcfspec as libvcfspec


def check_SV(vr):
    alt = vr.alts[0]
    if any(
        (re.fullmatch(f'<{x}(:.+)?>', alt) is not None)
        for x in libvcfspec.SV_ALTS
    ):
        # <DEL>, <INV>, ...
        return True
    elif (
        (libvcfspec.PAT_BND1.fullmatch(alt) is not None)
        or (libvcfspec.PAT_BND2.fullmatch(alt) is not None)
    ):
        return True
    else:
        return False


def check_cpgmet(vr):
    return vr.alts[0] == f'<{libvcfspec.CPGMET_ALT}>'


# sorting helpers

def get_vr_sortkey(chromdict):
    def sortkey(vr):
        return (chromdict.contigs.index(vr.contig), vr.pos, vr.ref) + vr.alts

    return sortkey


# applying vcfspec to vr

#def apply_vcfspec(vr, vcfspec):
#    vr.contig = vcfspec.chrom
#    vr.pos = vcfspec.pos
#    vr.ref = vcfspec.ref
#    vr.alts = vcfspec.alts
#    if not vcfspec.components.is_missing:
#        vcfspec.components.write(vr)


# vr reformatting

def merge(vr_list, vcfheader=None):
    """Intended to be run with variant records with identical vcfspecs.
    chrom, pos, id, ref, alt, qual, INFO/END is derived from the first vr.
    FILTER values are union'ed except PASS value.
    INFO and FORMAT values are chosen from the first variant record which
        contains the key.
    """

    if vcfheader is None:
        vcfheader = headerhandler.merge_vcfheaders([x.header for x in vr_list])

    new_vr = vcfheader.new_record()
    _copy_basic_attrs(vr_list[0], new_vr)

    for idx, vr in enumerate(vr_list):
        _copy_filter(vr, new_vr, skip_pass=True)
        _copy_info(vr, new_vr, dont_copy_if_exist=True)
        _copy_format(vr, new_vr, dont_copy_if_exist=True)

    # handle CallerInfo
    from handygenome.annotation.callerinfo import CallerInfo
    import handygenome.annotation.callerinfo as libcallerinfo

    if any(CallerInfo.get_annotkey() in vr.header.info.keys() for vr in vr_list):
        merged_callerinfo = libcallerinfo.get_merged_callerinfo(vr_list)
        merged_callerinfo.write(new_vr)

    return new_vr


def reheader(vr, vcfheader):
    """Return a new variant record created using the template
    header, filled with the contents of the input variant record.

    INFO or FORMAT keys or samples not included in the template header 
    are discarded.
    """

    new_vr = vcfheader.new_record()
    _copy_basic_attrs(vr, new_vr)
    _copy_filter(vr, new_vr, skip_pass=False)
    _copy_info(vr, new_vr, dont_copy_if_exist=False)
    _copy_format(vr, new_vr, dont_copy_if_exist=False)

    # amend missing values in new_vr which are
        # erroneously represented by .................
    new_vr_format_keys = tuple(new_vr.header.formats.keys())
    new_vr_samples = tuple(new_vr.samples.keys())
    for sampleid in new_vr_samples:
        for key in new_vr_format_keys:
            if new_vr.header.formats[key].number == 'G':
                continue
            if infoformat.check_NA_format(vr, sampleid, key):
                infoformat.set_NA_format(new_vr, sampleid, key)

    return new_vr


def rename(vr, samples):
    if len(samples) != len(vr.header.samples):
        raise Exception(
            f'The number of samples must be the same.')

    chromdict = refgenome.ChromDict.from_vcfheader(vr.header)
    new_header = initvcf.create_header(chromdict=chromdict, samples=samples, vcfheader=vr.header)
    new_vr = new_header.new_record()
    _copy_basic_attrs(vr, new_vr)
    _copy_filter(vr, new_vr, skip_pass=False)
    _copy_info(vr, new_vr, dont_copy_if_exist=False)

    for old_sampleid, new_sampleid in zip(vr.header.samples, samples):
        for key in vr.samples[old_sampleid]:
            if key == 'GT':
                new_vr.samples[new_sampleid].allele_indices = vr.samples[old_sampleid].allele_indices
                new_vr.samples[new_sampleid].phased = vr.samples[old_sampleid].phased
            else:
                infoformat.set_format(
                    vr=new_vr, 
                    sampleid=new_sampleid,
                    key=key, 
                    val=infoformat.get_format(vr, old_sampleid, key),
                )

    return new_vr


# vr reformatting helpers

def _copy_basic_attrs(old_vr, new_vr):
    for key in ('contig', 'pos', 'id', 'qual'):
        setattr(new_vr, key, getattr(old_vr, key))
    if old_vr.alleles is not None:
        # assigning None to vr.alleles raises an Exception
        new_vr.alleles = old_vr.alleles
    if old_vr.stop is not None:
        new_vr.stop = old_vr.stop


def _copy_filter(old_vr, new_vr, skip_pass=True):
    for filter_name in old_vr.filter:
        # check if new vr header defines the key
        if filter_name in new_vr.header.filters:
            if skip_pass:
                # copy filter only if it is not PASS
                if filter_name != 'PASS':
                    new_vr.filter.add(filter_name)
            else:
                new_vr.filter.add(filter_name)


def _copy_info(old_vr, new_vr, dont_copy_if_exist=True):
    for info_key in old_vr.info:
        # check if new vr header defines the key
        if info_key in new_vr.header.info:  
            # skip if Number == 'G'
            if new_vr.header.info[info_key].number == 'G':
                continue
            if dont_copy_if_exist:
                # copy the value only if the value is not already set
                if info_key not in new_vr.info:
                    infoformat.set_info(new_vr, info_key, 
                                        infoformat.get_info(old_vr, info_key))
            else:
                infoformat.set_info(new_vr, info_key, 
                                    infoformat.get_info(old_vr, info_key))


def _copy_format(old_vr, new_vr, dont_copy_if_exist=True):
    def assign_val(old_vr, new_vr, key, sampleid):
        if key == 'GT':
            new_vr.samples[sampleid].allele_indices = old_vr.samples[sampleid].allele_indices
            new_vr.samples[sampleid].phased = old_vr.samples[sampleid].phased
        else:
            infoformat.set_format(new_vr, sampleid, key, infoformat.get_format(old_vr, sampleid, key))

    new_vr_samples = tuple(new_vr.samples)
    for sampleid in old_vr.samples:
        # check if the new vr header defines the sample id of input vr
        if sampleid in new_vr_samples:
            for key in old_vr.samples[sampleid]:
                # check if new vr header defines the key
                if key in new_vr.header.formats:
                    # skip if Number == 'G'
                    if new_vr.header.formats[key].number == 'G':
                        continue

                    if dont_copy_if_exist:
                        # copy the value only if the value is not already set
                        if key not in new_vr.samples[sampleid]:
                            assign_val(old_vr, new_vr, key, sampleid)
                    else:
                        assign_val(old_vr, new_vr, key, sampleid)


# raw VCF string parsers
def parse_raw_info(vr):
    result = dict()
    infosp = str(vr).split('\t')[7].split(';')
    for x in infosp:
        xsp = x.split('=')
        if len(xsp) == 1:
            result[xsp[0]] = vr.info[xsp[0]]
        else:
            result[xsp[0]] = xsp[1]
    return result


# functions dealing with list of vr
    
def check_has_id(vr_list):
    """
    Args:
        vr_list: A list of pysam.VariantRecord objects
    """

    result = True
    for vr in vr_list:
        if vr.id is None:
            result = False
            break

    return result


def check_has_duplicate_pos_alleles(vr_list):
    """
    Args:
        vr_list: A list of pysam.VariantRecord objects
    """

    result = False
    IDset = set()
    for vr in vr_list:
        ID = (vr.contig, vr.pos, vr.alleles)
        if ID in IDset:
            result = True
            break
        else:
            IDset.add(ID)

    return result


def check_has_duplicate_id(vr_list):
    """
    Args:
        vr_list: A list of pysam.VariantRecord objects

    Returns:
        True when two or more vr have no ID.
    """

    has_duplicate_id = False
    IDset = set()
    for vr in vr_list:
        if vr.id in IDset:
            has_duplicate_id = True
            break
        else:
            IDset.add(vr.id)

    return has_duplicate_id


def check_has_mateid(vr_list):
    """
    Args:
        vr_list: A list of pysam.VariantRecord objects
    """

    has_mateid = True
    for vr in vr_list:
        if 'MATEID' not in vr.info.keys():
            has_mateid = False
            break
    
    return has_mateid


# VariantRecord pickling

def encode_vr(vr):
    return tuple(vr.samples), str(vr)


def decode_vr(vr_data):
    samples, vr_string = vr_data
    hdr = pysam.VariantHeader()
    for sid in samples:
        hdr.add_sample(sid)

    with tempfile.NamedTemporaryFile(delete=False, mode='wt') as tmpfile:
        tmpfile.write(str(hdr))
        tmpfile.write(vr_string)

    with pysam.VariantFile(tmpfile.name) as in_vcf:
        vr = next(in_vcf)

    os.remove(tmpfile.name)

    return vr





import importlib
import cyvcf2

top_package_name = __name__.split('.')[0]
module_common_params = importlib.import_module(top_package_name + '.common.params')
module_repeats = importlib.import_module(top_package_name + '.variantplus.repeats')
module_caller_format_check = importlib.import_module(top_package_name + '.svtools.caller_format_check')

variant_format_checker_dict, caller_list = moduel_caller_format_check.get_variant_format_checker_dict()


class VariantPlus:
	def __init__(
			self, 
			variant=None, 
			CHROM1=None, 
			POS1=None, 
			REF=None, 
			ALT=None,
			CHROM2=None, 
			POS2=None, 
			CT=None,
			fasta=None,
			):
		'''
		Argument requirements:
		1) 'variant' is supplied
		2) 'CHROM1', 'POS1', 'REF', 'ALT' are supplied
		'''
		set_basic_attributes(self, variant, CHROM1, POS1, REF, ALT, CHROM2, POS2, fasta)
		modify_REF(self)
		self.mttype = get_mttype(self.REF, self.ALT)
		set_repeat_list(self)


	def INFO_has_field(self, field):
		try:
			value = self.variant.INFO[field]
		except KeyError:
			return False
		else:
			if value == '.':
				return False
			else:
				return True


### set_basic_attributes

def set_basic_attributes(vp, variant, CHROM1, POS1, REF, ALT, CHROM2, POS2, fasta):
	if variant == None:
		set_basic_attributes_without_variant(vp, CHROM1, POS1, REF, ALT, CHROM2, POS2, fasta)
	else:
		set_basic_attributes_with_variant(vp, variant, fasta)


def set_basic_attributes_with_variant(vp, variant, fasta): # vp: variantplus
	vp.variant = variant
	vp.fasta = fasta

	vp.CHROM1 = variant.CHROM
	vp.POS1 = variant.POS
	vp.POS01 = vp.POS1 - 1
	vp.REF = variant.REF
	vp.ALT = variant.ALT[0]

	vp.CHROM2 = None
	vp.POS2 = None
	vp.POS02 = None


def set_basic_attributes_without_variant(vp, CHROM1, POS1, REF, ALT, CHROM2, POS2, fasta): # vp: variantplus
	vp.variant = None
	vp.fasta = fasta

	vp.CHROM1 = CHROM1
	vp.POS1 = POS1
	vp.POS01 = vp.POS1 - 1
	vp.REF = variant.REF
	vp.ALT = variant.ALT[0]

	vp.CHROM2 = CHROM2
	vp.POS2 = vp.POS2
	vp.POS02 = vp.POS2 - 1

### modify_REF

def modify_REF(vp):
	assert 'REF' in dir(vp)
	if vp.REF == 'N':
		assert vp.fasta != None
		vp.REF = vp.fasta.fetch(vp.CHROM1, vp.POS01, vp.POS01 + 1)

### set_repeat_list

def set_repeat_list(vp):
	vp.repeat_list = module_repeats.get_repeat_list(vp.CHROM1, vp.POS01, vp.fasta, pre=10, post=10)
	vp.relevant_repeat_list = module_repeats.get_relevant_repeat_list(
			vp.repeat_list, vp.CHROM1, vp.POS01, vp.REF, vp.ALT, mttype = vp.mttype, flanklen = 5,
			)

# get_mttype

def get_mttype(REF, ALT):
	if module_common_params.PAT_NUCLEOBASES.match(ALT): # ALT is [ACGTNacgtn]+
		if len(REF) == len(ALT):
			if len(REF) == 1:
				mttype = 'SNV'
			else:
				mttype = 'MNV'
		else:
			if len(REF) == 1:
				mttype = 'INS'
			elif len(ALT) == 1:
				mttype = 'DEL'
			else:
				mttype = 'CINDEL'
	else:
		mttype = 'SV'

	return mttype

# get_sv_format

def get_sv_format(variant):
	'''
	checks caller format of a variant line of a SV vcf
	'''

	true_formats = list()
	for caller, checker in variant_format_checker_dict.items():
		if checker(variant): # checker(variant) returns a boolean
			true_formats.append(caller)

	if len(true_formats) == 0:
		raise Exception(f'Input variant line format does not conform with any of the callers: {caller_list}')
	elif len(true_formats) > 1:
		raise Exception(f'Input variant line format conforms with more than one of these callers: {caller_list}')

	sv_format = true_formats[0]
	
	return sv_format




from handygenome.annotation.annotitem import AnnotItemInfoSingle


class CallerInfo(AnnotItemInfoSingle):
    meta = {
        "ID": "caller_info",
        "Number": "1",
        "Type": "String",
        "Description": "Information about which variant calling program called this variant.",
    }

    @classmethod
    def from_vr(cls, vr):
        return cls.from_vr_base(vr)

    def write(self, vr):
        self.write_base(vr)

    @classmethod
    def from_caller_names(cls, caller_names):
        result = cls.init_nonmissing()
        result['callers'] = tuple(caller_names)
        return result
    

def get_merged_callerinfo(vr_list):
    raw_dict = {'callers': set()}
    for vr in vr_list:
        if CallerInfo.get_annotkey() in vr.info.keys():
            unit_callerinfo = CallerInfo.from_vr(vr)
            raw_dict['callers'].update(unit_callerinfo['callers'])
    raw_dict['callers'] = tuple(raw_dict['callers'])

    result = CallerInfo.init_nonmissing()
    result.update(raw_dict)
    return result



DEFAULT_FLANKLEN_PARSIDE = 1
DEFAULT_FLANKLEN_BNDSIDE = 1
BND_DISTANCE_THRESHOLD = 1000


def make_alleleinfoitem_readpluspair(aiitem_bnd1_rp1, aiitem_bnd1_rp2,
                                     aiitem_bnd2_rp1, aiitem_bnd2_rp2, bnds):
    """Assumes that each rp supports at most 1 bnd (one of bnd1 or bnd2).
        This assumption is monitored in "update_alleleinfo_sv" method
        of ReadPlus class.
    """

    alleleinfoitem = {
        'ref_support_direct': False,
        'ref_support_indirect': False,
        'alt_support_direct': False,
        'alt_support_indirect': False,
        'other_support': False,
        'noninformative': False,
        }

    # other
    if any((aiitem_bnd1_rp1['other_support'], 
            aiitem_bnd2_rp1['other_support'],
            aiitem_bnd1_rp2['other_support'], 
            aiitem_bnd2_rp2['other_support'])):
        alleleinfoitem['other_support'] = True

    # alt
    if any((aiitem_bnd1_rp1['alt_support'], 
            aiitem_bnd2_rp1['alt_support'],
            aiitem_bnd1_rp2['alt_support'],
            aiitem_bnd2_rp2['alt_support'])):
        # direct alt support with read crossing
        alleleinfoitem['alt_support_direct'] = True
    elif (
            (
                aiitem_bnd1_rp1['facing_frombnd'] and 
                aiitem_bnd2_rp2['facing_frombnd']) or
            (
                aiitem_bnd1_rp2['facing_frombnd'] and 
                aiitem_bnd2_rp1['facing_frombnd'])):
        # the mate reads facing each other, on bndside of each breakend
        alleleinfoitem['alt_support_indirect'] = True

    # ref
    if not (alleleinfoitem['other_support'] or 
            alleleinfoitem['alt_support_direct'] or
            alleleinfoitem['alt_support_indirect']):
        # regarded not as ref-supporting when other- or alt- supporting
        if any((aiitem_bnd1_rp1['ref_support'], 
                aiitem_bnd2_rp1['ref_support'],
                aiitem_bnd1_rp2['ref_support'],
                aiitem_bnd2_rp2['ref_support'])):
            # direct ref support with read crossing
            alleleinfoitem['ref_support_direct'] = True
        elif (
                (
                    aiitem_bnd1_rp1['facing_frombnd'] and 
                    aiitem_bnd1_rp2['facing_frompar']) or
                (
                    aiitem_bnd1_rp2['facing_frombnd'] and 
                    aiitem_bnd1_rp1['facing_frompar']) or
                (
                    aiitem_bnd2_rp1['facing_frombnd'] and 
                    aiitem_bnd2_rp2['facing_frompar']) or
                (
                    aiitem_bnd2_rp2['facing_frombnd'] and 
                    aiitem_bnd2_rp1['facing_frompar'])):
            # the mate reads facing each other, one read on bndside of a 
                # breakend and the other read on parside of the same breakend
            alleleinfoitem['ref_support_indirect'] = True

    # noninformative
    if not any(v for (k, v) in alleleinfoitem.items() 
               if k != 'noninformative'):
        alleleinfoitem['noninformative'] = True

    return alleleinfoitem


def make_alleleinfoitem_readplus(bnds, is_bnd1, rp,
                                 flanklen_parside=DEFAULT_FLANKLEN_PARSIDE, 
                                 flanklen_bndside=DEFAULT_FLANKLEN_BNDSIDE):
    def handle_bndside_notonborder(rp, endis5, alleleinfoitem):
        if faces_border_frombnd(rp.read, endis5):
            if clipped_toward_border(rp.read, endis5):
                alleleinfoitem['noninformative'] = True
            else:
                alleleinfoitem['facing_frombnd'] = True
        else:
            alleleinfoitem['noninformative'] = True

    def handle_bndside_onborder_clipped(
            rp, pos_range0, endis5, alleleinfoitem, is_bnd1,
            flanklen_parside, bndside_flankrange0):
        queryseq_length = len(pos_range0) + flanklen_parside
        queryseq_beyond_bndsideflank = rp.get_seq_from_coord(
            start0=pos_range0.start, length=queryseq_length,
            forward=(not endis5))

        if len(queryseq_beyond_bndsideflank) < queryseq_length:
            alleleinfoitem['noninformative'] = True
        else:  # now flanking length conditions are fulfilled
            if rp.check_matches(bndside_flankrange0):
                seq_ref = bnds.get_seq_beyond_bnd_ref(is_bnd1=is_bnd1, 
                                                      length=flanklen_parside, 
                                                      with_border_seq=True)
                seq_alt = bnds.get_seq_beyond_bnd_alt(is_bnd1=is_bnd1, 
                                                      length=flanklen_parside,
                                                      with_border_seq=True)
                if queryseq_beyond_bndsideflank == seq_ref:
                    alleleinfoitem['ref_support'] = True
                elif queryseq_beyond_bndsideflank == seq_alt:
                    alleleinfoitem['alt_support'] = True
                else:
                    alleleinfoitem['other_support'] = True
            else:
                alleleinfoitem['other_support'] = True

    def handle_traverser(rp, pos_range0, endis5, flanklen_parside,
                         alleleinfoitem, bndside_flankrange0,
                         is_bnd1):
        queryseq_beyond_bnd = rp.get_seq_from_coord(
            start0=pos_range0.stop, length=flanklen_parside,
            forward=(not endis5))
        if len(queryseq_beyond_bnd) < flanklen_parside:
            alleleinfoitem['noninformative'] = True
        else:  # now flanking length conditions are fulfilled
            if (
                    rp.check_matches(bndside_flankrange0) and
                    rp.check_matches(pos_range0)):
                seq_beyond_bnd_ref = bnds.get_seq_beyond_bnd_ref(
                    is_bnd1=is_bnd1, length=flanklen_parside, 
                    with_border_seq=False)
                seq_beyond_bnd_alt = bnds.get_seq_beyond_bnd_alt(
                    is_bnd1=is_bnd1, length=flanklen_parside,
                    with_border_seq=False)

                if queryseq_beyond_bnd == seq_beyond_bnd_ref:
                    alleleinfoitem['ref_support'] = True
                elif queryseq_beyond_bnd == seq_beyond_bnd_alt:
                    alleleinfoitem['alt_support'] = True
                else:
                    alleleinfoitem['other_support'] = True
            else:
                alleleinfoitem['other_support'] = True

    # get parameters
    (chrom, pos_range0, endis5) = bnds.get_params(is_bnd1)
    bndside_flankrange0 = bnds.get_flank_range0(mode='bnd_dist', 
                                                is_bnd1=is_bnd1, 
                                                flanklen=flanklen_bndside)

    # initialize alleleinfoitem
    alleleinfoitem = {
        'noninformative': False,
        'ref_support': False,
        'alt_support': False,
        'other_support': False,
        'facing_frombnd': False,
        'facing_frompar': False,
        }

    # main
    if chrom == rp.read.reference_name:
        distance = rp.get_distance(pos_range0)
        if distance > BND_DISTANCE_THRESHOLD:
            alleleinfoitem['noninformative'] = True
        else:
            if on_bndside(rp.read, pos_range0, endis5):
                if on_bndside_notonborder(rp.read, pos_range0, endis5):
                    handle_bndside_notonborder(rp, endis5, alleleinfoitem)
                else:  # on bndside and on the border
                    if rp.check_spans(bndside_flankrange0):
                        if clipped_toward_border(rp.read, endis5):
                            handle_bndside_onborder_clipped(
                                rp, pos_range0, endis5, alleleinfoitem, 
                                is_bnd1, flanklen_parside, 
                                bndside_flankrange0)
                        else:
                            if faces_border_frombnd(rp.read, endis5):
                                alleleinfoitem['facing_frombnd'] = True
                            else:
                                alleleinfoitem['noninformative'] = True
                    else:
                        alleleinfoitem['noninformative'] = True

            elif on_parside(rp.read, pos_range0, endis5):
                if faces_border_frompar(rp.read, endis5):
                    alleleinfoitem['facing_frompar'] = True
                else:
                    alleleinfoitem['noninformative'] = True

            else:  # traverses the most advanced border
                if rp.check_spans(bndside_flankrange0):
                    handle_traverser(rp, pos_range0, endis5, flanklen_parside,
                                     alleleinfoitem, bndside_flankrange0,
                                     is_bnd1)
                else:
                    alleleinfoitem['noninformative'] = True
    else:
        alleleinfoitem['noninformative'] = True

    return alleleinfoitem


def on_bndside_notonborder(read, pos_range0, endis5):
    if endis5:
        return read.reference_start > pos_range0.start
    else:
        return read.reference_end <= pos_range0.start


def on_bndside(read, pos_range0, endis5):
    if endis5:
        return read.reference_start > pos_range0.stop
    else:
        return read.reference_end <= pos_range0.stop


# parside: partner side (named in conformity with the VCF spec)
def on_parside(read, pos_range0, endis5):
    if endis5:
        return read.reference_end <= pos_range0.stop + 1
    else:
        return read.reference_start >= pos_range0.stop


def clipped_toward_border(read, endis5):
    if endis5:
        return read.cigartuples[0][0] in (4, 5)
    else:
        return read.cigartuples[-1][0] in (4, 5)


def faces_border_frombnd(read, endis5):
    if endis5:
        return read.is_reverse
    else:
        return read.is_forward
   
    
def faces_border_frompar(read, endis5):
    if endis5:
        return read.is_forward
    else:
        return read.is_reverse










# helper functions

#def spans_bndsideflank(rp, bndsideflank_range0):
#    return (rp.ref_range0.start <= min(bndsideflank_range0) and
#            rp.ref_range0.stop > max(bndsideflank_range0))
#
#
#def spans_retracted_pos(rp, bndsideflank_range0):
#    return bndsideflank_range0.stop in rp.ref_range0
#
#
#def get_querypos0_beyond_bndsideflank(rp, bndsideflank_range0, endis5):
#    idx_retractedpos = rp.pairs_dict['refpos0'].index(
#        bndsideflank_range0.stop)
#
#    if endis5:
#        sl = slice(0, idx_retractedpos + 1)
#    else:
#        sl = slice(0, idx_retractedpos + 1)
#        querypos0_beyond_deep = rp.pairs_dict['querypos0'][idx_retractedpos:]
#
#    querypos0_beyond_deep = rp.pairs_dict['querypos0'][sl]
#    querypos0_beyond_deep = [x for x in querypos0_beyond_deep if x is not None]
#
#    return querypos0_beyond_deep
#
#
##def get_length_beyond_bndsideflank(pos_range0
#
#
#def spans_beyond_bnd(read, parflank_range0, endis5):
#    if endis5:
#        return read.reference_start <= parflank_range0[-1]
#    else:
#        return read.reference_end > parflank_range0[0]
#
#
#def spans_beyond_parflank(read, parflank_range0, endis5):
#    if endis5:
#        return read.reference_start <= parflank_range0[0]
#    else:
#        return read.reference_end > parflank_range0[-1]
#
#
#def get_length_otherside(read, pos_range0, endis5):
#    # assumes that the read does not span beyond the breakend
#    if endis5:
#        if read.cigartuples[0][0] == 4:
#            remaining_length = read.cigartuples[0][1]
#        else:
#            remaining_length = 0
#        length_thisside = pos_range0.stop - read.reference_start - 1
#    else:
#        if read.cigartuples[-1][0] == 4:
#            remaining_length = read.cigartuples[-1][1]
#        else:
#            remaining_length = 0
#        length_thisside = pos_range0.stop - read.reference_end
#
#    length_otherside = remaining_length - length_thisside
#
#    return length_otherside
#
#
#def clippedat_border(read, endis5):
#    if endis5:
#        return read.cigartuples[0][0] == 4
#    else:
#        return read.cigartuples[-1][0] == 4


#def get_supporting_bnds(rp):
#
#    def get_suppl_info(rp, SAitem):
#        """suppl_is5prime: 
#            Indicates whether the aligned portion in the supplementary 
#            alignment belongs to the 5-prime side (toward the first base) 
#            of the read.
#        suppl_isleft:
#            Indicates whether the aligned portion in the supplementary 
#            alignment belongs to the leftmost side, with regard to the
#            plus strand of the reference sequence.
#        """
#
#        if SAitem['is_forward']:
#            if (
#                    SAitem['cigartuples'][0][0] == 4 and 
#                    SAitem['cigartuples'][-1][0] == 0):
#                suppl_is5prime = False
#                suppl_isleft = False
#            elif (
#                    SAitem['cigartuples'][0][0] == 0 and 
#                    SAitem['cigartuples'][-1][0] == 4):
#                suppl_is5prime = True
#                suppl_isleft = True
#        else:
#            if (
#                    SAitem['cigartuples'][0][0] == 4 and 
#                    SAitem['cigartuples'][-1][0] == 0):
#                suppl_is5prime = True
#                suppl_isleft = False
#            elif (
#                    SAitem['cigartuples'][0][0] == 0 and 
#                    SAitem['cigartuples'][-1][0] == 4):
#                suppl_is5prime = False
#                suppl_isleft = True
#
#        return suppl_is5prime, suppl_isleft
#
#    def get_primary_clip_info(rp, suppl_is5prime):
#        """The softclip length in the primary alignment, 
#            corresponding to the supplementary alignment.
#        """
#
#        if suppl_is5prime:
#            clip_cigartuple = (rp.read.cigartuples[0]
#                               if rp.read.is_forward else
#                               rp.read.cigartuples[-1])
#        else:
#            clip_cigartuple = (rp.read.cigartuples[-1]
#                               if rp.read.is_forward else
#                               rp.read.cigartuples[0])
#
#        if clip_cigartuple[0] != 4:
#            raise Exception(
#                f'The cigarop of the primary read on the SA side is not '
#                f'a softclip.')
#
#        primary_cliplen = clip_cigartuple[1]
#
#        return primary_cliplen
#
#    # main
#    if rp.SAlist is None:
#        return None
#    
#    for SAitem in rp.SAlist:
#        suppl_is5prime, suppl_isleft = get_suppl_info(rp, SAitem)
#        primary_cliplen = get_primary_cliplen(rp, suppl_is5prime)
                
                
        
        
        



import os
import collections
import random
import itertools
import functools
import multiprocessing
import operator

import Bio.SeqUtils
import numpy as np
import pandas as pd
import pyranges as pr
import matplotlib.pyplot as plt
import seaborn as sns
import scipy
#from scipy.stats.mstats import winsorize

import handygenome.common as common
import handygenome.pyranges_helper as pyranges_helper
import handygenome.cnv.mosdepth as libmosdepth
import handygenome.cnv.sequenza_handler as sequenza_handler
import handygenome.cnv.gcfraction as gcfraction
import handygenome.assemblyspec as libassemblyspec


MALE_HAPLOID_CHROMS = ('X', 'Y', 'chrX', 'chrY')
DEFAULT_CNT_WEIGHT = 5


class CCFInfo(
    collections.namedtuple('CCFInfo', ('CNm', 'ccf', 'mutated_allele'))
):
    pass


def get_normal_mean_ploidy(refver, is_female):
    """Only supports hg19 or hg38"""
    if is_female:
        return 2
    else:
        chromdict = common.DEFAULT_CHROMDICTS[refver]
        N_region_gr = libassemblyspec.get_N_region_gr(refver)
        par_gr = libassemblyspec.get_par_gr(refver)

        autosomal_chroms = [
            x for x in chromdict.contigs 
            if (
                (common.RE_PATS['assembled_chromosome'] is not None)
                and (x not in MALE_HAPLOID_CHROMS)
            )
        ]  # 1, 2, ...,
        X_chrom, Y_chrom = chromdict.XY_names
            # X, Y or chrX, chrY

        CNs = list()
        lengths = list()
        # autosomes
        for chrom in autosomal_chroms:
            CNs.append(2)
            N_length = N_region_gr[N_region_gr.Chromosome == chrom].lengths().sum()
            lengths.append(chromdict[chrom] - N_length)
        # X - PAR
        CNs.append(2)
        PAR_X_length = par_gr[par_gr.Chromosome == X_chrom].lengths().sum()
        lengths.append(PAR_X_length)
        # X - non-PAR
        CNs.append(1)
        X_N_length = N_region_gr[N_region_gr.Chromosome == X_chrom].lengths().sum()
        lengths.append(chromdict[X_chrom] - X_N_length - PAR_X_length)
        # Y
        CNs.append(1)
        PAR_Y_length = par_gr[par_gr.Chromosome == Y_chrom].lengths().sum()
        Y_N_length = N_region_gr[N_region_gr.Chromosome == Y_chrom].lengths().sum()
        lengths.append(chromdict[Y_chrom] - Y_N_length - PAR_Y_length)

        return np.average(CNs, weights=lengths)


def check_haploid(is_female, chrom):
    return (not is_female) and (chrom in MALE_HAPLOID_CHROMS)


def get_CNn_list(refver, is_female, chroms, start0s, end0s):
    """Assumes that N-masked regions or Y-PAR is not given as arguments"""
    chromdict = common.DEFAULT_CHROMDICTS[refver]
    X_chrom, Y_chrom = chromdict.XY_names
    if refver in libassemblyspec.PARS:
        pars = libassemblyspec.PARS[refver]
        par1x = pars['PAR1_X']
        par2x = pars['PAR2_X']
        par_unavailable = False
    else:
        par_unavailable = True

    is_haploid_dict = {
        chrom: check_haploid(is_female, chrom)
        for chrom in set(chroms)
    }

    # main
    result = list()
    for chrom, start0, end0 in zip(chroms, start0s, end0s):
        if is_haploid_dict[chrom]:
            if chrom == X_chrom:
                if par_unavailable:  # Assumes CNn is globally 1 if PAR is not available
                    CNn = 1
                else:
                    if any(
                        common.check_overlaps(start0, end0, par[1], par[2])
                        for par in (par1x, par2x)
                    ):  # any overlap with PAR is regarded as CNn == 2
                        CNn = 2
                    else:
                        CNn = 1
            elif chrom == Y_chrom:
                CNn = 1
            else:
                raise Exception(f'Non-XY haploid chromosome')
        else:
            CNn = 2

        result.append(CNn)

    return result


##############################################


def theoretical_depth_ratio(
    CNt, cellularity, tumor_mean_ploidy, CNn, normal_mean_ploidy, 
    tumor_avg_depth_ratio=1, normal_avg_depth_ratio=1,
):
    """Args:
        CNn, normal_ploidy must be nonzero
    """
    tumor_upper_term = (CNt * cellularity) + (CNn * (1 - cellularity))
    tumor_lower_term = (tumor_mean_ploidy * cellularity) + (normal_mean_ploidy * (1 - cellularity))
    tumor_norm_depth = tumor_avg_depth_ratio * (tumor_upper_term / tumor_lower_term)

    #normal_upper_term = CNn
    #normal_lower_term = normal_mean_ploidy
    #normal_norm_depth = normal_avg_depth_ratio * (normal_upper_term / normal_lower_term)
    normal_norm_depth = normal_avg_depth_ratio * (CNn / normal_mean_ploidy)

    return tumor_norm_depth / normal_norm_depth


def inverse_theoretical_depth_ratio(
    depth_ratio, cellularity, tumor_mean_ploidy, CNn, normal_mean_ploidy, 
    tumor_avg_depth_ratio=1, normal_avg_depth_ratio=1,
):
    """Args:
        cellularity must be nonzero
    Returns:
        CNt estimate

    """
    '''
    <Induction>
    tumor_upper_term = (CNt * cellularity) + (CNn * (1 - cellularity))
    tumor_lower_term = (tumor_mean_ploidy * cellularity) + (normal_mean_ploidy * (1 - cellularity))
    tumor_norm_depth = tumor_avg_depth_ratio * (tumor_upper_term / tumor_lower_term)
    normal_norm_depth = normal_avg_depth_ratio * (CNn / normal_mean_ploidy)
    depth_ratio = tumor_norm_depth / normal_norm_depth
    ---------------------------------
    tumor_norm_depth = depth_ratio * normal_norm_depth
    tumor_avg_depth_ratio * (tumor_upper_term / tumor_lower_term) = depth_ratio * normal_norm_depth
    tumor_upper_term = (depth_ratio * normal_norm_depth * tumor_lower_term) / tumor_avg_depth_ratio
    (CNt * cellularity) + (CNn * (1 - cellularity)) = (depth_ratio * normal_norm_depth * tumor_lower_term) / tumor_avg_depth_ratio
    CNt * cellularity = ((depth_ratio * normal_norm_depth * tumor_lower_term) / tumor_avg_depth_ratio) - (CNn * (1 - cellularity))
    CNt = (
        ((depth_ratio * normal_norm_depth * tumor_lower_term) / tumor_avg_depth_ratio) - (CNn * (1 - cellularity))
    ) / cellularity
    '''
    normal_norm_depth = normal_avg_depth_ratio * (CNn / normal_mean_ploidy)
    tumor_lower_term = (tumor_mean_ploidy * cellularity) + (normal_mean_ploidy * (1 - cellularity))
    CNt = (
        ((depth_ratio * normal_norm_depth * tumor_lower_term) / tumor_avg_depth_ratio) - (CNn * (1 - cellularity))
    ) / cellularity
    return CNt


@functools.cache
def delta_depth_ratio(
    cellularity, tumor_mean_ploidy, CNn, normal_mean_ploidy, 
    tumor_avg_depth_ratio=1, normal_avg_depth_ratio=1,
):
    """Returns:
        Difference of depth ratio caused by 1 CNt change
    """
    '''
    <Induction>
    tumor_upper_term = (CNt * cellularity) + (CNn * (1 - cellularity))
    tumor_lower_term = (tumor_mean_ploidy * cellularity) + (normal_mean_ploidy * (1 - cellularity))
    tumor_norm_depth = tumor_avg_depth_ratio * (tumor_upper_term / tumor_lower_term)
    normal_norm_depth = normal_avg_depth_ratio * (CNn / normal_mean_ploidy)
    depth_ratio = tumor_norm_depth / normal_norm_depth
    ------------------------------------------
    depth_ratio = (tumor_avg_depth_ratio * (tumor_upper_term / tumor_lower_term)) / normal_norm_depth
    depth_ratio = tumor_upper_term * (tumor_avg_depth_ratio / (tumor_lower_term * normal_norm_depth))
    depth_ratio = (
        ((CNt * cellularity) + (CNn * (1 - cellularity))) 
        * 
        (tumor_avg_depth_ratio / (tumor_lower_term * normal_norm_depth))
    )
    delta_depth_ratio = (
        ((CNt_1 * cellularity) - (CNt_2 * cellularity)) 
        * 
        (tumor_avg_depth_ratio / (tumor_lower_term * normal_norm_depth))
    )
    delta_depth_ratio = (
        cellularity
        * 
        (tumor_avg_depth_ratio / (tumor_lower_term * normal_norm_depth))
    )
    delta_depth_ratio = (cellularity * tumor_avg_depth_ratio) / (tumor_lower_term * normal_norm_depth)
    '''
    tumor_lower_term = (tumor_mean_ploidy * cellularity) + (normal_mean_ploidy * (1 - cellularity))
    normal_norm_depth = normal_avg_depth_ratio * (CNn / normal_mean_ploidy)
    delta_depth_ratio = (cellularity * tumor_avg_depth_ratio) / (tumor_lower_term * normal_norm_depth)
    return delta_depth_ratio


def theoretical_depth_ratio_sequenza(CNt, cellularity, ploidy, CNn=2, normal_ploidy=2, avg_depth_ratio=1):
    """Args:
        CNn, normal_ploidy must be nonzero
    """
    cellularity_term = ((CNt / CNn) * cellularity) + (1 - cellularity)
    ploidy_term = ((ploidy / normal_ploidy) * cellularity) + (1 - cellularity)
    return avg_depth_ratio * (cellularity_term / ploidy_term)


def inverse_theoretical_depth_ratio_sequenza(depth_ratio, cellularity, ploidy, CNn=2, normal_ploidy=2, avg_depth_ratio=1):
    """Args:
        cellularity must be nonzero
    Returns:
        CNt estimate
    """
    ploidy_term = ((ploidy / normal_ploidy) * cellularity) + (1 - cellularity)
    # depth_ratio * ploidy_term == avg_depth_ratio * cellularity_term
    return (((depth_ratio * ploidy_term) / avg_depth_ratio) - (1 - cellularity)) * (CNn / cellularity)


@functools.cache
def delta_depth_ratio_sequenza(cellularity, ploidy, CNn=2, normal_ploidy=2, avg_depth_ratio=1):
    ploidy_term = ((ploidy / normal_ploidy) * cellularity) + (1 - cellularity)
    return (avg_depth_ratio / ploidy_term) * (cellularity / CNn)


def theoretical_baf(CNt, B, cellularity, CNn):
    if CNn <= 1:
        return None
    else:
        numerator = (B * cellularity) + (1 - cellularity)
        denominator = (CNt * cellularity) + (CNn * (1 - cellularity))
        return numerator / denominator


def inverse_theoretical_baf(baf, cellularity, CNt, CNn):
    """Args:
        cellularity must be nonzero
    Returns:
        B estimate
    """
    '''
    numerator = (B * cellularity) + (1 - cellularity)
    denominator = (CNt * cellularity) + (CNn * (1 - cellularity))
    baf = numerator / denominator

    baf = ((B * cellularity) + (1 - cellularity)) / denominator
    (B * cellularity) + (1 - cellularity) = baf * denominator
    B * cellularity = (baf * denominator) - (1 - cellularity)
    B = ((baf * denominator) - (1 - cellularity)) / cellularity
    '''
    denominator = (CNt * cellularity) + (CNn * (1 - cellularity))
    B = ((baf * denominator) - (1 - cellularity)) / cellularity
    return B


def get_B(CNt, cellularity, baf):
    B_estimate = inverse_theoretical_baf(baf, cellularity, CNt)
    if B_estimate <= 0:
        B = 0
        theo_baf = theoretical_baf(CNt, B, cellularity)
        if theo_baf is None:
            B = None
            diff = None
        else:
            diff = abs(baf - theo_baf)
    else:
        B_candidate_upper = int(np.ceil(B_estimate))
        B_candidate_lower = int(np.floor(B_estimate))

        theo_baf_upper = theoretical_baf(CNt, B_candidate_upper, cellularity)
        theo_baf_lower = theoretical_baf(CNt, B_candidate_lower, cellularity)

        if (theo_baf_upper is None) and (theo_baf_lower is None):
            B = None
            diff = None
        elif (theo_baf_upper is not None) and (theo_baf_lower is None):
            B = B_candidate_upper
            diff = abs(baf - theo_baf_upper)
        elif (theo_baf_upper is None) and (theo_baf_lower is not None):
            B = B_candidate_lower
            diff = abs(baf - theo_baf_lower)
        else:
            diff_upper = abs(theo_baf_upper - baf)
            diff_lower = abs(theo_baf_lower - baf)
            if diff_upper <= diff_lower:
                B = B_candidate_upper
                diff = diff_upper
            else:
                B = B_candidate_lower
                diff = diff_lower

    return B, diff


def get_CN_from_cp(cellularity, ploidy, depth_ratio, baf, CNt_weight):
    def save_cache(CNt_candidate, B_cache, segfit_score_cache, baf_diff_cache):
        B, baf_diff = get_B(CNt_candidate, cellularity, baf)
        if B is None:
            # drop this CNt candidate if B cannot be calculated
            return

        ratio_diff = abs(theoretical_depth_ratio(CNt_candidate, cellularity, ploidy) - depth_ratio)
        B_cache[CNt_candidate] = B
        baf_diff_cache[CNt_candidate] = baf_diff
        segfit_score_cache[CNt_candidate] = ratio_diff * CNt_weight + baf_diff

    # get initial CNt candidates
    CNt_estimate = inverse_theoretical_depth_ratio(depth_ratio, cellularity, ploidy)
    if CNt_estimate <= 0:
        initial_candidate_upper = 0
        initial_candidate_lower = None
    else:
        initial_candidate_upper = int(np.ceil(CNt_estimate))
        initial_candidate_lower = int(np.floor(CNt_estimate))

    # set caches
    B_cache = dict()
    segfit_score_cache = dict()
    baf_diff_cache = dict()

    delta_ratio = delta_depth_ratio(cellularity, ploidy) * CNt_weight

    # upper
    save_cache(initial_candidate_upper, B_cache, segfit_score_cache, baf_diff_cache)
    if initial_candidate_upper in baf_diff_cache:
        n_further_candidates = int(baf_diff_cache[initial_candidate_upper] / delta_ratio)
        for offset in range(n_further_candidates):
            CNt_candidate = initial_candidate_upper + (offset + 1)
            save_cache(CNt_candidate, B_cache, segfit_score_cache, baf_diff_cache)
    # lower
    if initial_candidate_lower is not None:
        save_cache(initial_candidate_lower, B_cache, segfit_score_cache, baf_diff_cache)
        if initial_candidate_lower in baf_diff_cache:
            try:
                n_further_candidates = int(baf_diff_cache[initial_candidate_lower] / delta_ratio)
            except:
                print(cellularity, ploidy, depth_ratio, baf)
                print(baf, baf_diff_cache[initial_candidate_lower], initial_candidate_lower, delta_ratio)
                raise

            for offset in range(n_further_candidates):
                CNt_candidate = initial_candidate_lower - (offset + 1)
                if CNt_candidate < 0:
                    break
                save_cache(CNt_candidate, B_cache, segfit_score_cache, baf_diff_cache)

    # result
    if len(segfit_score_cache) == 0:
        return None, None, None
    else:
        CNt, segfit_score = min(segfit_score_cache.items(), key=operator.itemgetter(1))
        B = B_cache[CNt]
        #selected_score = segfit_score_cache[CNt]

        return CNt, B, segfit_score


def get_CN_from_cp_wo_baf(cellularity, ploidy, depth_ratio, CNt_weight):
    CNt_estimate = inverse_theoretical_depth_ratio(depth_ratio, cellularity, ploidy)
    if CNt_estimate <= 0:
        CNt = 0
        diff = abs(theoretical_depth_ratio(CNt, cellularity, ploidy) - depth_ratio)
        segfit_score = diff * CNt_weight
    else:
        upper_candidate = int(np.ceil(CNt_estimate))
        lower_candidate = int(np.floor(CNt_estimate))

        upper_diff = abs(theoretical_depth_ratio(upper_candidate, cellularity, ploidy) - depth_ratio)
        lower_diff = abs(theoretical_depth_ratio(lower_candidate, cellularity, ploidy) - depth_ratio)
        if upper_diff <= lower_diff:
            CNt = upper_candidate
            segfit_score = upper_diff * CNt_weight
        else:
            CNt = lower_candidate
            segfit_score = lower_diff * CNt_weight

    return CNt, None, segfit_score  # B is None


def calc_cp_score(segment_df, cellularity, ploidy, is_female, CNt_weight):
    segfit_score_sum = 0
    start0_list = list()
    end0_list = list()
    CNt_list = list()
    for idx, row in segment_df.iterrows():
        if np.isnan(row['depth_mean']):
            continue

        if check_haploid(is_female, row['Chromosome']) or np.isnan(row['baf_mean']):
            CNt, B, segfit_score = get_CN_from_cp_wo_baf(cellularity, ploidy, row['depth_mean'], CNt_weight)
        else:
            CNt, B, segfit_score = get_CN_from_cp(cellularity, ploidy, row['depth_mean'], row['baf_mean'], CNt_weight)

        if segfit_score is not None:
            #segfit_score_sum += segfit_score
            segfit_score_sum -= segfit_score  # to make greater score better

        start0_list.append(row['Start'])
        end0_list.append(row['End'])
        CNt_list.append(CNt)

    seg_mean_ploidy = _get_mean_ploidy_from_values(start0_list, end0_list, CNt_list)
    ploidyfit_diff = abs(seg_mean_ploidy - ploidy)

    return {'segfit_score': segfit_score_sum, 'ploidyfit_diff': ploidyfit_diff}


def get_cp_score_dict(segment_df, is_female, CNt_weight=DEFAULT_CNT_WEIGHT, nproc=None):
    c_candidates = np.round(np.arange(0.01, 1.00, 0.01), 2)  # 0 and 1 are excluded
    p_candidates = np.round(np.arange(0.5, 7.1, 0.1), 1)  # 0.5 to 7.0
    cp_pairs = tuple(itertools.product(c_candidates, p_candidates))
    with multiprocessing.Pool(nproc) as pool:
        scorelist = pool.starmap(
            calc_cp_score, 
            ((segment_df, x[0], x[1], is_female, CNt_weight) for x in cp_pairs),
        )
    cp_scores = dict(zip(cp_pairs, scorelist))
    return cp_scores


def get_cpscore_peaks(cp_scores):
    c_list = sorted(set((x[0] for x in cp_scores.keys())))
    p_list = sorted(set((x[1] for x in cp_scores.keys())))
    cpscore_df_source = dict()
    for c in c_list:
        cpscore_df_source[c] = [cp_scores[(c, p)] for p in p_list]
    cpscore_df = pd.DataFrame.from_dict(cpscore_df_source)
    cpscore_df.index = p_list
    cpscore_df = -1 * cpscore_df  # to be run with get_fitresult_peaks function

    peaks = sequenza_handler.get_fitresult_peaks(cpscore_df)
    for x in peaks:
        x['score'] = -1 * x['lpp'] 
        del x['lpp']

    return peaks


def add_CN_to_segment(segment_df, cellularity, ploidy, is_female, CNt_weight=DEFAULT_CNT_WEIGHT):
    CNt_list = list()
    B_list = list()
    score_list = list()
    for idx, row in segment_df.iterrows():
        if np.isnan(row['depth_mean']):
            CNt, B, score = None, None, None
        else:
            if check_haploid(is_female, row['Chromosome']) or np.isnan(row['baf_mean']):
                CNt, B, score = get_CN_from_cp_wo_baf(cellularity, ploidy, row['depth_mean'], CNt_weight)
            else:
                CNt, B, score = get_CN_from_cp(cellularity, ploidy, row['depth_mean'], row['baf_mean'], CNt_weight)

        if CNt is None:
            CNt = np.nan
        if B is None:
            B = np.nan
        if score is None:
            score = np.nan

        CNt_list.append(CNt)
        B_list.append(B)
        score_list.append(score)

    return segment_df.assign(**{'CNt': CNt_list, 'B': B_list, 'score': score_list})


def add_theoreticals_to_segment(segment_df, cellularity, ploidy, is_female):
    theo_depthratio_list = list()
    theo_baf_list = list()
    for idx, row in segment_df.iterrows():
        if np.isnan(row['CNt']):
            theo_depthratio = np.nan
            theo_baf = np.nan
        else:
            if check_haploid(is_female, row['Chromosome']):
                CNn = 1
            else:
                CNn = 2

            # depthratio
            theo_depthratio = theoretical_depth_ratio(row['CNt'], cellularity, ploidy, CNn=CNn)
            # baf
            if np.isnan(row['B']):
                theo_baf = np.nan
            else:
                theo_baf = theoretical_baf(row['CNt'], row['B'], cellularity, CNn=CNn)
                if theo_baf is None:
                    theo_baf = np.nan

        theo_depthratio_list.append(theo_depthratio)
        theo_baf_list.append(theo_baf)

    return segment_df.assign(
        **{'predicted_depth_ratio': theo_depthratio_list, 'predicted_baf': theo_baf_list}
    )
    

###################################


def _get_mean_ploidy_from_values(start0_list, end0_list, CNt_list):
    lengths = pd.Series(end0_list) - pd.Series(start0_list)
    return np.average(CNt_list, weights=lengths)


def get_mean_ploidy(seg_df):
    seg_lengths = seg_df.End - seg_df.Start
    numerator = (seg_df.CNt * seg_lengths).sum()
    denominator = seg_lengths.sum()
    return numerator / denominator


def theoretical_somatic_vaf(CNt, CNm, cellularity, ccf, CNn=2):
    """Assumes no overlapping subclonal CNV.
    Args:
        CNm: Mutated copy number
        CNn: Germline copy number (should be 1 in haploid area such as male sex chromosome)
        ccf: Cancer cell fraction
    """
    numerator = CNm * cellularity * ccf
    denominator = CNt * cellularity + CNn * (1 - cellularity)
    return numerator / denominator


def ccf_from_params(vaf, CNt, CNm, cellularity, CNn):
    """Reverse of 'theoretical_somatic_vaf'"""
    numerator = vaf * (CNt * cellularity + CNn * (1 - cellularity))
    denominator = CNm * cellularity
    return numerator / denominator


def CNm_likelihood_difference(observed_vaf, CNt, CNm, cellularity, CNn):
    expected_vaf = theoretical_somatic_vaf(
        CNt=CNt, CNm=CNm, cellularity=cellularity, ccf=1, CNn=CNn,
    )
    return -abs(expected_vaf - observed_vaf)


def CNm_likelihood_binom(total_read, var_read, CNt, CNm, cellularity, CNn):
    expected_vaf = theoretical_somatic_vaf(
        CNt=CNt, CNm=CNm, cellularity=cellularity, ccf=1, CNn=CNn,
    )
    return scipy.stats.binom.pmf(var_read, total_read, expected_vaf)


def get_ccf_CNm(vaf, cellularity, CNt, CNB=None, likelihood='diff', total_read=None, var_read=None, CNn=2):
    """Notes:
        - Assumes no overlapping subclonal CNV
        - If subclonal, mutation may reside in only one copy. CNm must be 1.
        - If clonal, CNm can be any value between 1 and max(CNA, CNB).
    Args:
        CNn: Should be 1 in haploid area such as male sex chromosome
        CNB: None when haploid
    """
    assert likelihood in ('diff', 'binom')

    # NA cases
    if (
        CNt == 0 or
        np.isnan(vaf)
    ):
        return CCFInfo(None, None, None)

    # get CNA
    if CNB is None:
        CNA = None
    else:
        CNA = CNt - CNB
    # max_subclonal_vaf results when CNm == 1 and ccf == 1
    max_subclonal_vaf = theoretical_somatic_vaf(
        CNt=CNt, CNm=1, cellularity=cellularity, ccf=1, CNn=CNn,
    )

    if vaf >= max_subclonal_vaf:
        # Assumes clonal mutation
        ccf = 1
        max_CNm = CNt if (CNB is None) else max(CNA, CNB)
        CNm_candidates = list(range(1, max_CNm + 1))
        if likelihood == 'diff':
            CNm = max(
                CNm_candidates, 
                key=(lambda x: CNm_likelihood_difference(
                    observed_vaf=vaf, CNt=CNt, CNm=x, cellularity=cellularity, CNn=CNn,
                )),
            )
        elif likelihood == 'binom':
            CNm = max(
                CNm_candidates, 
                key=(lambda x: CNm_likelihood_binom(
                    total_read=total_read, 
                    var_read=var_read, 
                    CNt=CNt, 
                    CNm=x, 
                    cellularity=cellularity, 
                    CNn=CNn,
                )),
            )
    else:
        # Assumes subclonal mutation
        CNm = 1
        ccf = ccf_from_params(vaf=vaf, CNt=CNt, CNm=CNm, cellularity=cellularity, CNn=CNn)

    # get which allele is mutated
    if CNB is None:
        mutated_allele = None
    else:
        if CNA == CNB:
            mutated_allele = None
        else:
            if CNm > min(CNA, CNB):
                if CNA > CNB:
                    mutated_allele = 'A'
                else:
                    mutated_allele = 'B'
            else:
                mutated_allele = None

    return CCFInfo(CNm, ccf, mutated_allele)


#################################
# depth df processing functions #
#################################

def get_gc_breaks(n_bin):
    breaks = np.linspace(0, 1, (n_bin + 1), endpoint=True)
    breaks[0] = -0.001
    return breaks


def remove_outliers(depth_df, trim_limits=(0.05, 0.05), lower_cutoff=None, upper_cutoff=None):
    """Absolute cutoff-based trimming is done first, then proportional trimming is later
    """
    # remove nan
    #selector = np.repeat(True, depth_df.shape[0])
    selector = ~np.isnan(depth_df['mean_depth'])

    # trim by absolute cutoff
    if lower_cutoff is not None:
        selector = np.logical_and(selector, depth_df['mean_depth'] >= lower_cutoff)
    if upper_cutoff is not None:
        selector = np.logical_and(selector, depth_df['mean_depth'] <= upper_cutoff)

    depth_df = depth_df.loc[selector, :]

    # trim by proportion
    trimmed_depth = scipy.stats.mstats.trim(
        depth_df['mean_depth'], limits=trim_limits, relative=True,
    )
    depth_df = depth_df.loc[~trimmed_depth.mask, :]

    return depth_df


def get_depth_df_from_file(chroms, start0s, end0s, depth_file_path):
    preset_depth_df = pd.read_csv(
        depth_file_path, 
        sep='\t', 
        names=('Chromosome', 'Start', 'End', 'depth'), 
        dtype={'Chromosome': str, 'Start': int, 'End': int, 'depth': float},
    )
    preset_depth_gr = pr.PyRanges(preset_depth_df)
    result_gr = pr.PyRanges(chromosomes=chroms, starts=start0s, ends=end0s)
    result_gr = pyranges_helper.join(result_gr, preset_depth_gr, how='left', merge='weighted_mean', as_gr=True)
    return result_gr


# NOT USED
def get_gc_depth_data(depth_df, n_bin):
    assert 'GC' in depth_df.columns

    breaks = get_gc_breaks(n_bin)
    cutresult = pd.cut(depth_df['GC'], bins=breaks)

    gcbin_average_depths = dict()
    for intv, subdf in depth_df.groupby(cutresult, axis=0):
        # since "cutresult" is a categorical series, iteration over groupby object yields all bins including empty ones
        if subdf.shape[0] == 0:
            gcbin_average_depths[intv] = np.nan
        else:
            gcbin_average_depths[intv] = np.average(subdf['mean_depth'], weights=(subdf['End'] - subdf['Start']))
                # weighted average depth for the given GC bin

    gcbin_average_depths = pd.Series(gcbin_average_depths)
    gcbin_norm_average_depths = gcbin_average_depths / gcbin_average_depths.mean(skipna=True)

    return gcbin_average_depths, gcbin_norm_average_depths, cutresult


def get_gcbin_data(depth_df, gc_breaks):
    """This does not apply weights when calculating averages. Suitable for even-sized bins"""
    # cut by gc bins
    cutresult = pd.cut(depth_df['GC'], bins=gc_breaks)
    # make results
    gcbin_depth_data = {intv: list() for intv in cutresult.dtype.categories}
    for intv, depth in zip(cutresult, depth_df['mean_depth']):
        gcbin_depth_data[intv].append(depth)

    gcbin_average_depths = {
        intv: (np.nan if len(data) == 0 else np.mean(data))
        for intv, data in gcbin_depth_data.items()
    }
    gcbin_average_depths = pd.Series(gcbin_average_depths)
    gcbin_norm_average_depths = gcbin_average_depths / gcbin_average_depths.mean(skipna=True)

    return gcbin_depth_data, gcbin_average_depths, gcbin_norm_average_depths, cutresult


def postprocess_depth_df(
    depth_df, 
    *,
    fasta=None, 
    refver=None,
    binsize=None,
    gc_window=None,

    nan_lower_cutoff=None,
    nan_upper_cutoff=None,

    gcdata_trim_limits=(0.05, 0.05),
    gcdata_lower_cutoff=None,
    gcdata_upper_cutoff=None,

    n_gcbin=100, 
    as_gr=True, 
):
    """Args:
        fasta, refver, binsize: Only used for gc fraction calculation.
            Allowed argument usage: "fasta" only OR "refver" and "binsize" only

    Returns:
        postprocessed df & gcbin average depth dict
        Adds 4 columns:
            - GC
            - norm_mean_depth
            - gc_corrected_mean_depth
            - sequenza_style_norm_mean_depth
    """
    # sanity check
    if not (
        (
            (fasta is not None )
            and (refver is None)
            and (binsize is None)
        )
        or (
            (fasta is None )
            and (refver is not None)
            and (binsize is not None)
        )
    ):
        raise Exception(
            f'Allowed argument usage for GC fraction annotation: '
            f'1) use "fasta", do NOT use "refver" and "binsize" ; '
            f'2) use "refver" and "binsize", do NOT use "fasta"'
        )

    assert 'mean_depth' in depth_df.columns, f'"depth_df" must include a column named "mean_depth"'

    # replace outliers with np.nan
    if nan_lower_cutoff is not None:
        depth_df.loc[depth_df['mean_depth'] < nan_lower_cutoff, ['mean_depth']] = np.nan
    if nan_upper_cutoff is not None:
        depth_df.loc[depth_df['mean_depth'] >= nan_upper_cutoff, ['mean_depth']] = np.nan

    # GC
    gcfraction.add_gc_calculating(depth_df, fasta, window=gc_window)

    # norm_mean_depth
    depth_df_wo_nan = depth_df.loc[~np.isnan(depth_df['mean_depth']), :]
    global_mean_depth = np.average(depth_df_wo_nan['mean_depth'], weights=(depth_df_wo_nan['End'] - depth_df_wo_nan['Start']))
    depth_df.insert(depth_df.shape[1], 'norm_mean_depth', (depth_df['mean_depth'] / global_mean_depth).array)

    # get gc depth data
    gc_breaks = get_gc_breaks(n_gcbin)
    trimmed_depth_df = remove_outliers(depth_df, trim_limits=gcdata_trim_limits, lower_cutoff=gcdata_lower_cutoff, upper_cutoff=gcdata_upper_cutoff)
    gcbin_depth_data, gcbin_average_depths, gcbin_norm_average_depths, cutresult = get_gcbin_data(trimmed_depth_df, gc_breaks)

    # gc_corrected_mean_depth
    cutresult = pd.cut(depth_df['GC'], bins=gc_breaks)
    gcbin_norm_average_depths_selected = gcbin_norm_average_depths.loc[cutresult]
    depth_df.insert(depth_df.shape[1], 'gc_corrected_mean_depth', (depth_df['mean_depth'].array / gcbin_norm_average_depths_selected.array))

    # sequenza_style_norm_mean_depth
    gcbin_average_depths_selected = gcbin_average_depths.loc[cutresult]
    depth_df.insert(depth_df.shape[1], 'sequenza_style_norm_mean_depth', (depth_df['mean_depth'].array / gcbin_average_depths_selected.array))

    if as_gr:
        return pr.PyRanges(depth_df, int64=False), gcbin_average_depths
    else:
        return depth_df, gcbin_average_depths


def get_processed_depth_df(bam_path, fasta, region_gr, gc_window=None, outlier_cutoffs='wgs', donot_subset_bam=False, as_gr=False):
    # set params
    assert isinstance(outlier_cutoffs, (tuple, list)) or outlier_cutoffs in ('wgs', 'panel')

    if outlier_cutoffs == 'wgs':
        lower_cutoff = 0
        upper_cutoff = 2000
    elif outlier_cutoffs == 'panel':
        lower_cutoff = 50
        upper_cutoff = np.inf
    else:
        lower_cutoff, upper_cutoff = outlier_cutoffs

    # run mosdepth
    depth_df, _ = libmosdepth.run_mosdepth(
        bam_path,
        t=8, 
        region_gr=region_gr, 
        donot_subset_bam=donot_subset_bam, 
        as_gr=False, 
        load_perbase=False,
    )

    # postprocess
    depth_gr, gcbin_average_depths = postprocess_depth_df(
        depth_df, 

        fasta=fasta, 
        gc_window=gc_window,

        gcdata_trim_limits=(0.05, 0.05),
        gcdata_lower_cutoff=lower_cutoff,
        gcdata_upper_cutoff=upper_cutoff,
        nan_lower_cutoff=lower_cutoff,
        nan_upper_cutoff=upper_cutoff,

        n_gcbin=100,
        as_gr=as_gr, 
    )

    return depth_gr, gcbin_average_depths


def make_depth_ratio_df(tumor_depth_df, normal_depth_df, as_gr=False):
    # sanity check
    assert isinstance(tumor_depth_df, pd.DataFrame) and isinstance(normal_depth_df, pd.DataFrame)
    assert all(
        {'mean_depth', 'GC', 'norm_mean_depth', 'gc_corrected_mean_depth', 'sequenza_style_norm_mean_depth'}.issubset(df.columns)
        for df in (tumor_depth_df, normal_depth_df)
    )

    compared_columns = ['Chromosome', 'Start', 'End', 'GC']
    if not (tumor_depth_df.loc[:, compared_columns] == normal_depth_df.loc[:, compared_columns]).to_numpy().all():
        raise Exception(f'The columns {compared_columns} of the two input dataframes must be identical.')

    # main
    result = tumor_depth_df.loc[:, compared_columns].copy()
    result['depth_ratio_sequenzastyle'] = tumor_depth_df['sequenza_style_norm_mean_depth'] / normal_depth_df['sequenza_style_norm_mean_depth']
    result['depth_ratio_mystyle'] = tumor_depth_df['gc_corrected_mean_depth'] / normal_depth_df['gc_corrected_mean_depth']
    for colname in ('mean_depth', 'norm_mean_depth', 'gc_corrected_mean_depth', 'sequenza_style_norm_mean_depth'):
        result[f'tumor_{colname}'] = tumor_depth_df.loc[:, colname]
        result[f'normal_{colname}'] = normal_depth_df.loc[:, colname]

    if as_gr:
        return pr.PyRanges(result)
    else:
        return result


#########################################
# plotting GCfrac - depth relationships #
#########################################

def arghandler_ylims(ydata, ylims):
    if ylims is None:
        return None
    else:
        return (
            min(ylims[0], min(ydata)), 
            max(ylims[1], max(ydata)),
        )


def plot_gc_depth(
    fasta, bam_path, 
    gc_window=50, 
    xwidth=1000, 
    chrom=None,
    start0=None,
    end0=None,
    return_gr=True, 
    depth_ylims=None, 
    gc_ylims=None,
    alpha=0.6,
    linewidth=1,
):
    """Intended to be used with short regions (~10kb)
    GC fraction value is calculated for every single base, with a window "gc_window" wide.
    """
    pad_left = int(gc_window / 2)
    if gc_window % 2 == 0:
        pad_right = pad_left
    else:
        pad_right = pad_left + 1    

    if any(x is None for x in [chrom, start0, end0]):
        chrom = random.choice(fasta.references[:24])
        chromlen = fasta.lengths[fasta.references.index(chrom)]
        start0 = random.choice(range(pad_left, (chromlen - xwidth - pad_right) + 1))
        end0 = start0 + xwidth

    fetch_start0 = start0 - pad_left
    fetch_end0 = end0 + pad_right
    
    ### GC ###
    seq = fasta.fetch(chrom, fetch_start0, fetch_end0)
    x = list()
    y_gc = list()
    for idx in range(0, end0 - start0):
        subseq = seq[idx:(idx + gc_window)]
        gcfrac = Bio.SeqUtils.gc_fraction(subseq)
        x.append(start0 + idx)
        y_gc.append(gcfrac)

    # modify ylims
    gc_ylims = arghandler_ylims(y_gc, gc_ylims)
    
    # plot
    fig, ax_gc = plt.subplots(figsize=(20, 5))
    if gc_ylims is not None:
        ax_gc.set_ylim(*gc_ylims)
    ax_gc.set_title(f'{chrom}:{start0 + 1:,}-{end0:,}')
    ax_gc.plot(
        x, y_gc, 
        color='green', marker=None, linestyle='-', markersize=1, linewidth=linewidth, alpha=alpha
    )
    
    ### depth ###
    region_gr = pr.from_dict({'Chromosome': [chrom], 'Start': [start0], 'End': [end0]})
    gr, gr_perbase = libmosdepth.run_mosdepth(
        bam_path, 
        use_median=False,
        region_gr=region_gr, 
        as_gr=True, 
        load_perbase=True,
    )
    y_depth = list(gr_perbase.depth)

    # modify ylims
    depth_ylims = arghandler_ylims(y_depth, depth_ylims)

    # plot
    ax_depth = ax_gc.twinx()
    if depth_ylims is not None:
        ax_depth.set_ylim(*depth_ylims)
    ax_depth.plot(
        x, y_depth, 
        color='black', marker=None, linestyle='-', markersize=1, linewidth=linewidth, alpha=alpha
    )

    # return
    if return_gr:
        gr_perbase.GC = y_gc
        return gr_perbase


def plot_gc_distribution(
    depth_df, 

    n_gcbin=100,

    trim_limits=(0.05, 0.05),
    lower_cutoff=None,
    upper_cutoff=None,

    histogram_mode='2d',
    scatter_ylims=None,
    heatmap_ylims=None, 
    heatmap_vmax=None,
):
    def make_average_scatter(gcbin_average_depths, ylims=None):
        x = list()
        y = list()
        for key, val in gcbin_average_depths.items():
            x.append(key.mid)
            y.append(np.mean(val))

        ylims = arghandler_ylims(y, ylims)

        fig, ax = plt.subplots(figsize=(10, 5))
        if ylims is not None:
            ax.set_ylim(*ylims)
        ax.set_xlabel('GC fraction')
        ax.set_ylabel('mean depth')
        ax.plot(x, y, 'ko', markersize=2)
        return fig

    def make_heatmap(gcbin_depth_data, ylims=None, vmax=None):
        ydata = list(itertools.chain.from_iterable(gcbin_depth_data.values()))
        if ylims is None:
            ylims = (min(ydata), max(ydata)) 
        else:
            ylims = arghandler_ylims(ydata, ylims)

        # set breaks
        breaks = np.linspace(ylims[0], ylims[1], 100, endpoint=True)

        # get histogram
        data = dict()
        for gc_interval, val in gcbin_depth_data.items():
            data[gc_interval] = np.histogram(val, bins=breaks, density=True)[0]
        df_plot = pd.DataFrame.from_dict(data)
        df_plot.index = pd.IntervalIndex.from_breaks(np.round(breaks, 1))

        # draw
        fig, ax = plt.subplots(figsize=(15, 10))
        sns.heatmap(df_plot, ax=ax, vmax=vmax)
        ax.set_xlabel('GC fraction')
        ax.set_ylabel('mean depth')
        return fig

    def make_heatmap_hist2d(depth_df, n_gcbin, ylims=None, vmax=None):
        # set breaks
        if ylims is None:
            ylims = (min(depth_df['mean_depth']), max(depth_df['mean_depth'])) 
        else:
            ylims = arghandler_ylims(depth_df['mean_depth'], ylims)
        depth_breaks = np.linspace(ylims[0], ylims[1], 100, endpoint=True)
        #gc_breaks = np.linspace(0, 1, 101, endpoint=True)
        gc_breaks = get_gc_breaks(n_gcbin)

        # get histogram
        H, xedges, yedges = np.histogram2d(
            x=depth_df['mean_depth'], y=depth_df['GC'], bins=(depth_breaks, gc_breaks), density=True,
        )
        plot_df = pd.DataFrame(
            H, 
            index=pd.IntervalIndex.from_breaks(np.round(xedges, 1)), 
            columns=pd.IntervalIndex.from_breaks(np.round(yedges, 2)),
        )
      
        # draw
        fig, ax = plt.subplots(figsize=(15, 10))
        sns.heatmap(plot_df, ax=ax, vmax=vmax)
        ax.set_xlabel('GC fraction')
        ax.set_ylabel('mean depth')
        return fig

    # main
    assert histogram_mode in ('1d', '2d')

    depth_df = remove_outliers(depth_df, trim_limits=trim_limits, lower_cutoff=lower_cutoff, upper_cutoff=upper_cutoff)
    gc_breaks = get_gc_breaks(n_gcbin)
    gcbin_depth_data, gcbin_average_depths, gcbin_norm_average_depths, cutresult = get_gcbin_data(depth_df, gc_breaks)
    fig_avg_scatter = make_average_scatter(gcbin_average_depths, ylims=scatter_ylims)
    if histogram_mode == '2d':
        fig_heatmap = make_heatmap_hist2d(depth_df, n_gcbin, ylims=heatmap_ylims, vmax=heatmap_vmax)
    elif histogram_mode == '1d':
        fig_heatmap = make_heatmap(gcbin_depth_data, ylims=heatmap_ylims, vmax=heatmap_vmax)

    return fig_avg_scatter, fig_heatmap


